from dataclasses import dataclass
from typing import Any


@dataclass
class Event:
    """
    A dataclass that defines an emitted event

    Attributes:
        name (str): The event name, used for parsing and listening for events
        content (Any): The content of the event, can contain anything
    """

    name: str
    content: Any

    def to_dict(self) -> dict:
        """

        Returns:
            A dictionary version of the object, received by vars()
        """
        return vars(self)



"""
Basic Usage:

>>> from avents import parse
>>> from avents import listen
>>> from avents import Event
>>>
>>> async def some_event_emitter():
>>>     await parse(Event("event", "content"))
>>>
>>> @listen("event")
>>> async def some_listener(event: Event):
>>>     print(event)
>>>

Advanced Usage:

>>> from avents import parse
>>> from avents import listen
>>> from avents import Event
>>> from avents import EventListener
>>> from avents import BaseEventType
>>>
>>>
>>> class CustomEventType(BaseEventType):
>>>     CUSTOM_EVENT: str = "custom event"
>>>     SUB_EVENT: str = "sub event"
>>>
>>>
>>> class CustomListener(EventListener):
>>>     ...
>>>
>>>
>>> async def example_event_emitter():
>>>     await parse(Event("custom event",
>>>             {"name": "sub event",
>>>              "content": "sub event content"})
>>>     )
>>>
>>> @listen(CustomEventType.CUSTOM_EVENT)
>>> async def base_parser(emitted_event: Event):
>>>     await CustomListener.parse(Event(**emitted_event.content))
>>>
>>> @CustomListener.listen(CustomEventType.SUB_EVENT)
>>> async def sub_parser(sub_event: Event):
>>>     print(sub_event)
>>>

"""

from avents.listener import EventListener
from avents.type import BaseEventType
from avents.event import Event

listen = EventListener().listen
parse = EventListener().parse

__all__ = [
    "listen",
    "parse",
    "EventListener",
    "BaseEventType",
    "Event"
]



from enum import Enum
from enum import EnumMeta


class EventTypeMeta(EnumMeta):
    """
    A metaclass added to help with iteration of enum values
    """

    def __contains__(cls, item):
        """
        Allows iteration of an enum to see if a value exists like so:

        >>> class SomeEnum(Enum, metaclass=EventTypeMeta):
        >>>     some_element = 5
        >>>
        >>> print(5 in SomeEnum)
        >>> > True

        Args:
            item: Some elemt input for comparison

        Returns:
            A bool representing whether or not item is in the enum
        """
        try:
            cls(item)
        except ValueError:
            return False
        return True


class BaseEventType(Enum, metaclass=EventTypeMeta):
    """
    A base Enum type included that should be inherited to define listenable events outside of raw strings

    For example:

    >>> class MyEventType(BaseEventType):
    >>>     EVENT_1: str = "EVENT 1"
    >>>     EVENT_2: str = "EVENT 2"

    This is useful when passing into EventListener's listen decorator argument on which events to emit
    """

    def __str__(self):
        return self.value



import asyncio
from avents.event import Event


class EventListener:
    """
    A class that implements a listen decorator and parser.

    Can be inherited from to generate custom listeners like so:

    >>> class CustomListener(EventListener):
    >>>     ...

    A constructor is not necessary on child classes, but you can still do the following:

    >>> class CustomListener(EventListener):
    >>>     def __init__(self):
    >>>         super().__init__()

    Inheriting EventListener to create a custom listener allows sub parsing of events, for example:

    >>> class CustomListener(EventListener):
    >>>     ...
    >>>
    >>>
    >>> @EventListener.listen("BASE EVENT")
    >>> async def base_event_handler(event: Event):
    >>>     # Where event.content contains a dict defining another event class, e.g.
    >>>     # {
    >>>     #   "name": "EVENT NAME",
    >>>     #   "content" "some event content"
    >>>     # }
    >>>     await CustomListener.parse(Event(**event.content))
    >>>
    >>> @CustomListener.listen("EVENT NAME")
    >>> async def some_custom_event(event: Event):
    >>>     print(event.content)
    """

    def __init__(self):
        """This allows this class to be inheritable to extend events to other classes"""
        cls = self.__class__

        try:
            _listening_events = getattr(cls, "_listening_events")
            if not _listening_events.get(cls):
                _listening_events[cls] = {}

        except AttributeError:
            setattr(cls, "_listening_events", {})

    @classmethod
    def listen(cls, *events: str):
        """
        A decorator that can be used to wrap coroutines (async functions) such that they listen for event emissions

        This registers each listening coroutine to the listeners parent class

        For example:

        >>> @EventListener.listen("some event")
        >>> async def some_event(event: Event):
        >>>     print(event)

        Listen can also handle multiple events:

        >>> @EventListener.listen("some event", "another event", "yet another event")
        >>> async def event_listener(event: Event):
        >>>     print(event)

        Args:
            event (str): The name of the event (the first argument passed to the Event class)

        Returns:
            The decorator wrap
        """
        def decorator(function):
            for event in events:
                event = str(event)
                _listening_events: dict = cls().__getattribute__("_listening_events")


                if not _listening_events[cls].get(event):
                    _listening_events[cls][event] = []
                _listening_events[cls][event].append(function)

        return decorator

    @classmethod
    async def parse(cls, event: Event):
        """
        A coroutine that is used to parse events

        For Example:

        >>> async def some_func():
        >>>     for i in range(2):
        >>>         await EventListener().parse(Event(name=str(i), content=f"Content-{i}"))

        This would call functions wrapped with the listen decorator for each proper event name

        Args:
            event (Event): An object defining events to be emitted to listening coroutines
        """
        _listening_events: dict = cls().__getattribute__("_listening_events")

        if events := _listening_events[cls].get(event.name, None):
            tasks = []
            for listener in events:
                tasks.append(listener(event))
            await asyncio.gather(*tasks)



#!/Users/sam/PycharmProjects/avents/venv/bin/python

from setuptools import setup

setup(
    name="avents",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    version=open("VERSION", "r").read(),
    packages=["avents"],
    python_requires=">=3.9",
    license_files=("LICENSE",),
    include_package_data=True,
    zip_safe=True,
    include_dirs=False,
    url="https://gitlab.cloud-technology.io/Open-Source/async-events",
    project_urls={
        "Documentation": "https://gitlab.cloud-technology.io/Open-Source/async-events/docs/",
        "Tracker": "https://gitlab.cloud-technology.io/Open-Source/async-events/issues/",
    },
    license="BSD 3-Clause",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    author="Price Hiller",
    author_email="philler3138@gmail.com",
    description="An asynchronous event library",
)
